// MassBuilderSaveTool
// Copyright (C) 2021-2022 Guillaume Jacquemin
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

#include <Corrade/Containers/String.h>

#include "../BinaryReader.h"
#include "../BinaryWriter.h"
#include "../PropertySerialiser.h"
#include "../../Logger/Logger.h"

#include "ArrayPropertySerialiser.h"

auto ArrayPropertySerialiser::deserialiseProperty(Containers::StringView name, Containers::StringView type,
                                                  UnsignedLong value_length, BinaryReader& reader,
                                                  PropertySerialiser& serialiser) -> UnrealPropertyBase::ptr
{
    Containers::String item_type;
    if(!reader.readUEString(item_type)) {
        LOG_ERROR_FORMAT("Couldn't read the item type of array property {}.", name);
        return nullptr;
    }

    char terminator;
    if(!reader.readChar(terminator) || terminator != '\0') {
        LOG_ERROR_FORMAT("Couldn't read a null byte in array property {}.", name);
        return nullptr;
    }

    UnsignedInt item_count;
    if(!reader.readUnsignedInt(item_count)) {
        LOG_ERROR_FORMAT("Couldn't read array property {}'s item count.", name);
        return nullptr;
    }

    auto prop = Containers::pointer<ArrayProperty>();
    prop->itemType = std::move(item_type);
    prop->items = serialiser.readSet(reader, prop->itemType, item_count);

    return prop;
}

auto ArrayPropertySerialiser::serialiseProperty(UnrealPropertyBase::ptr& prop, UnsignedLong& bytes_written,
                                                BinaryWriter& writer, PropertySerialiser& serialiser) -> bool
{
    auto array_prop = dynamic_cast<ArrayProperty*>(prop.get());
    if(!array_prop) {
        LOG_ERROR("The property is not a valid array property.");
        return false;
    }

    writer.writeUEStringToArray(array_prop->itemType);
    writer.writeValueToArray<char>('\0');
    bytes_written += writer.writeValueToArray<UnsignedInt>(UnsignedInt(array_prop->items.size()));

    UnsignedLong start_pos = writer.arrayPosition();
    UnsignedLong dummy_bytes_written = 0;
    bool ret = serialiser.writeSet(array_prop->items, array_prop->itemType, dummy_bytes_written, writer);
    bytes_written += writer.arrayPosition() - start_pos;

    return ret;
}
